#!/usr/bin/env perl
# the document root directory relative to the server's URL including /~user, if
# any.  For example, if the full URL path to index.html is
# http://foo.bar.net/~user/index.html, $doc_root should be set to ".".  Leave
# it blank if you can run u.cgi outside of the cgi-bin directory.
my $doc_root = "";

################################################################################
# Uniqki:	Unique Wiki <http://uniqki.ath.cx>
# Author:	Huidae Cho
# Since:	May 23, 2007
#
# Copyright (C) 2007 Huidae Cho <http://geni.ath.cx>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

################################################################################
use strict;
use vars qw(
	$U_CGI $HTTP_BASE $DOC_BASE

	$HOSTING

	$HTTP_COOKIE $HTTP_HOST $SERVER_NAME $SCRIPT_NAME $PATH_INFO
	$QUERY_STRING $REQUEST_METHOD $CONTENT_LENGTH

	$BODY_ONLY $USER $PAGE $FILE $VERSION $TITLE $TEXT

	$TZ $PW $PW_EXPIRATION $TPL $CSS $CHARSET

	$RSS_TITLE $RSS_DESCRIPTION

	$HEADER $FOOTER

	$WIKI_HEADER $WIKI_FOOTER $WIKI_NEW_PAGE $WIKI_UPLOAD

	$wiki $pre_text $line $end $post_text

	$protocol $protocol_char $protocol_punct $image_ext $text $block
	$reg_i0 $reg_i @reg @reg_sub $toc %h_i $h_top $h_prev $pre $table
	$li_i @li @li_p $p
);

umask 022;

################################################################################
# CGI variables
$HTTP_COOKIE = $ENV{HTTP_COOKIE};
$HTTP_HOST = $ENV{HTTP_HOST};
$SERVER_NAME = $ENV{SERVER_NAME};
$SCRIPT_NAME = $ENV{SCRIPT_NAME};
$PATH_INFO = $ENV{PATH_INFO};
$QUERY_STRING = $ENV{QUERY_STRING};
$REQUEST_METHOD = $ENV{REQUEST_METHOD};
$CONTENT_LENGTH = $ENV{CONTENT_LENGTH};

################################################################################
# useful variables
$U_CGI = $SCRIPT_NAME;
$HTTP_BASE = "http://$HTTP_HOST";
$DOC_BASE = "$HTTP_BASE$U_CGI"; $DOC_BASE =~ s#/[^/]*$##;
$PAGE = substr $PATH_INFO, 1; $PAGE =~ s#/.*$##;
$PAGE =~ s#\.(?:html|txt|txt\.v)$##;
$FILE = $PATH_INFO; $FILE =~ s#^/[^/]+##; $FILE =~ s#^/##;

################################################################################
# awardspace.com free web hosting
$HOSTING = "";
if(-d "/home/www/$SERVER_NAME"){
	$HOSTING = "awardspace";
	$_ = "/home/www/$SERVER_NAME$SCRIPT_NAME";
	s#/[^/]*$##;
	chdir $_;
}
if($doc_root eq ""){
	$U_CGI =~ s#^.*/##;
}else{
	($_ = $U_CGI) =~ s#/[^/]+$##;
	s#^/~[^/]+##;
	if($_ eq ""){
		chdir $doc_root;
	}else{
		s#^/##; s#[^/]+#..#g;
		chdir "$_/$doc_root";
	}
	$DOC_BASE = $HTTP_BASE.($U_CGI =~ m#(^/~[^/]+)# ? $1:"")."/$doc_root";
}

################################################################################
# config
do "u.cfg";
if($TZ eq ""){
	$TZ = "GMT";
}elsif($TZ =~ m/^gmt([+-])([0-9]+)$/i){
	$TZ = "GMT".($1 eq "+" ? "-" : "+").$2;
}
$PW = "u.pw" if($PW eq "");
$PW_EXPIRATION = 10 if($PW_EXPIRATION !~ m/^[0-9]+$/);
$TPL = "u.tpl" if($TPL eq "");
$CSS = "$TPL/uniqki.css" if($CSS eq "");
$CHARSET = "utf-8" if($CHARSET eq "");

$ENV{TZ} = $TZ; eval "use POSIX; POSIX::tzset();";

################################################################################
# initialization
$USER = "";
$BODY_ONLY = 0;
my $rebuild = 0;

################################################################################
# subroutines
#-------------------------------------------------------------------------------
# md5_hex routine for password checking
eval "use Digest::MD5 qw(md5_hex);";
if($@){
	sub md5_hex{
		my $str = shift;
		$str =~ s/'/'"'"'/g;
		return substr `printf '%s' '$str' | if echo "" | md5 > /dev/null 2>&1; then md5; else md5sum; fi`, 0, 32;
	}
}

sub lcs{
	my ($c0, $c1) = @_;
	my @lcs;
	my $s;
	for($s=0; $s<=$#$c0&&$s<=$#$c1; $s++){
		last if($$c0[$s] ne $$c1[$s]);
	}
	my ($e0, $e1);
	for($e0=$#$c0,$e1=$#$c1; $e0>$s&&$e1>$s; $e0--,$e1--){
		last if($$c0[$e0] ne $$c1[$e1]);
	}
	my ($m, $n);
	for($m=$s; $m<=$e0; $m++){
		for($n=$s; $n<=$e1; $n++){
			if($$c0[$m] eq $$c1[$n]){
				if($m && $n){
					$lcs[$m][$n] = $lcs[$m-1][$n-1] + 1;
				}else{
					$lcs[$m][$n] = 1;
				}
			}elsif($m && $n && $lcs[$m][$n-1]+0 > $lcs[$m-1][$n]+0){
				$lcs[$m][$n] = $lcs[$m][$n-1] + 0;
			}else{
				if($m){
					$lcs[$m][$n] = $lcs[$m-1][$n] + 0;
				}else{
					$lcs[$m][$n] = 0;
				}
			}
		}
	}
	my $i = $lcs[$e0][$e1];
	my @delta;
	$delta[$i] = ($e0+1).",".($e1+1);
	for($m=$e0,$n=$e1; $i>0&&$m>=$s&&$n>=$s; $m--,$n--){
		if($$c0[$m] eq $$c1[$n]){
			$delta[--$i] = "$m,$n";
		}elsif($lcs[$m][$n-1] > $lcs[$m-1][$n]){
			$m++;
		}else{
			$n++;
		}
	}
	return ($s, @delta);
}

sub diff{
	# split ignores trailing separators:
	# split /:/, "::1:2::" => ("", "", "1", "2")
	my @line0 = split /\n/, "$_[0]\x01";
	my @line1 = split /\n/, "$_[1]\x01";
	# discard \x01.
	$#line0--; $#line1--;
	my ($s, @delta) = lcs(\@line0, \@line1);
	my ($m, $n) = ($s, $s);
	my $diff = "";
	for(my $i=0; $i<=$#delta; $i++,$m++,$n++){
		my ($x, $y) = split /,/, $delta[$i];
		if($x > $m){
			for(; $m<$x; $m++){
				$diff .= "-$m\n";
			}
		}
		if($y > $n){
			for(; $n<$y; $n++){
				$diff .= "+$m $line1[$n]\n";
			}
		}
	}
	return $diff;
}

sub patch{
	my @line0 = split /\n/, "$_[0]\x01";
	my @lined = split /\n/, $_[1];
	$#line0--;
	my $line0p = "";
	for(my $i=0; $i<=$#lined; $i++){
		if($lined[$i] =~ m/^\+/){
			my $p = index $lined[$i], " ";
			my $j = substr $lined[$i], 1, $p-1;
			my $l = substr $lined[$i], $p+1;
			if($j){
				$line0[$j-1] .= "\n$l";
			}else{
				$line0p .= "$l\n";
			}
		}else{
			$line0[substr $lined[$i], 1] = "\x01";
		}
	}
	$line0[0] = "$line0p$line0[0]";
	my $str = join "\n", @line0;
	$str .= "\n";
	$str =~ s/\x01\n//g;
	return $str;
}

sub save{
	local *FH;
	my ($PAGE, $TEXT) = @_;
	if(-f "$PAGE.txt"){
		local $/ = undef;
		my $version = 0;
		my $txtv = "";
		if(open FH, "$PAGE.txt.v"){
			$version = <FH>;
			$txtv = $version.<FH>;
			close FH;
		}
		open FH, "$PAGE.txt";
		my $text = <FH>;
		close FH;
		my $diff = diff($TEXT, $text);
		if($diff ne ""){
			$version++;
			$txtv = "$version\n$diff\x01\n$txtv";
			open FH, ">$PAGE.txt.v";
			print FH $txtv;
			close FH;
		}
	}
	if(open FH, ">$PAGE.txt"){
		print FH $TEXT;
		close FH;
		$rebuild = 1;
	}
}

sub get_version{
	local *FH;
	my $PAGE = shift;
	my $version = 0;
	if(open FH, "$PAGE.txt.v"){
		$version = <FH> + 1;
		close FH;
		die "file corrupted!" unless(-f "$PAGE.txt");
	}elsif(-f "$PAGE.txt"){
		$version = 1;
	}
	return $version;
}

sub mylock{
	local *FH;
	my $timeout = 60;
	my $i = 0;
	while(-f "$_[0].lock"){
		die "mylock failed!" if(++$i > $timeout);
		sleep 1;
	}
	die "mylock failed!" unless(open FH, ">$_[0].lock");
	close FH;
}

sub myunlock{
	unlink "$_[0].lock";
}

sub make_html{
	my $preview = $#_ > 0;

	local $PAGE = shift;
	my $wikiedit = shift;
	my $text = shift;
	my $uploaded = shift;

	local *FH;
	my $txt;

	local $TITLE = "";
	local $TEXT;
	local $wiki;

	if($preview){
		open FH, ">", \$txt;
		print FH ($wikiedit ? "#!wiki\n" : "")."$text\n";
		close FH;

		$TEXT = uniqki(\$txt);
	}else{
		$txt = "$PAGE.txt";
		$TEXT = uniqki($txt);
	}

	chomp $TEXT;

	$TITLE = $PAGE if($TITLE eq "");

	local $BODY_ONLY = 1 if($preview);

	my $html;
	open FH, ">", \$html; my $fh = select FH;
	do "$TPL/".($wikiedit || ($wiki && !$preview) ? "wiki" : "")."display.tpl";
	close FH; select $fh;
	$html =~ s/\r//g;

	if($preview){
		$html =~ s/<!-- # -->.*\n//g;

		my $output;
		open FH, ">", \$output; my $fh = select FH;

		$TITLE = "Uniqki Preview: $TITLE";
		$BODY_ONLY = 0; do "$TPL/header.tpl";
		print $uploaded; print $html;
		$TEXT = $text;
		$TEXT =~ s/&/&amp;/g; $TEXT =~ s/</&lt;/g; $TEXT =~ s/>/&gt;/g;
		$BODY_ONLY = 1; do "$TPL/".($wikiedit ? "wiki" : "")."edit.tpl";
		$BODY_ONLY = 0; do "$TPL/footer.tpl";

		close FH; select $fh;

		return $output;
	}

	mylock("$PAGE.html");
	open FH, ">$PAGE.html";
	print FH $html;
	close FH;
	chmod 0755, "$PAGE.html" if($HOSTING eq "awardspace");
	myunlock("$PAGE.html");
}

sub rmrf{
	local *DH;
	foreach(@_){
		if(-f $_){
			unlink $_;
		}elsif(-d $_){
			my $dir = $_;
			opendir DH, $dir;
			my @i = map {"$dir/$_"} grep !/^\.{1,2}$/, readdir DH;
			closedir DH;
			rmrf(@i);
			rmdir $dir;
		}
	}
}

sub rmdirp{
	foreach(@_){
		my $dir = $_;
		while(index($dir, "/") >= 0){
			rmdir $dir;
			$dir =~ s#/[^/]*$##;
		}
		rmdir $dir;
	}
}

sub find{
	my $file = shift;
	if(-d $file){
		local *DH;
		my @list;
		opendir DH, $file;
		foreach(sort readdir DH){
			next if($_ eq "." || $_ eq "..");
			push @list, find("$file/$_");
		}
		closedir DH;
		return @list;
	}elsif(-f $file){
		return $file;
	}
	return;
}

sub get_var{
	my (%var, $v);
	foreach(split /&/, $QUERY_STRING){
		m/^([^=]*)=(.*)$/;
		$v = $1; $var{$v} = $2;
		$var{$v} =~ y/+/ /; $var{$v} =~ s/%0D//g;
		$var{$v} =~ s/%(..)/pack "c", hex($1)/eg;
	}
	my $boundary = <STDIN>;
	if("-" eq substr $boundary, 0, 1){
		read STDIN, my $content, $CONTENT_LENGTH-length($boundary);
		$content = $boundary.$content;
		(my $b = $boundary) =~ s/\r\n$//;
		$content =~ s/^$boundary//;
		$content =~ s/\r\n$b--\r\n$//;
		foreach(split /\r\n$boundary/, $content){
			my ($header, $body) = m/(.*?)\r\n\r\n(.*)$/s;
			$header =~ m/ name="(.*?)"/;
			my $name = $1;
			if($header =~ m/ filename="(.*?)"/){
				($var{$name} = $1) =~ s#^.*[/\\]##;
				$var{"$name="} = $body;
			}else{
				$body =~ s/\r//g;
				$var{$name} = $body;
			}
		}
	}else{
		foreach(split /&/, $boundary){
			m/^([^=]*)=(.*)$/;
			$v = $1; $var{$v} = $2;
			$var{$v} =~ y/+/ /; $var{$v} =~ s/%0D//g;
			$var{$v} =~ s/%(..)/pack "c", hex($1)/eg;
		}
	}
	return %var;
}

sub backup{
	eval "use Archive::Zip;";
	return if($@);

	my $page = shift;
	my $zip = Archive::Zip->new();
	my $file;
	if($page eq ""){
		$file = "uniqki.zip";
		$zip->addTree(".");
	}else{
		$file = "$page.zip";
		foreach("txt", "txt.v", "html"){
			$zip->addFile("$page.$_") if(-f "$page.$_");
		}
		$zip->addTree($page, $page) if(-d $page);
	}
	print "Content-Type: application/zip\n";
	print "Content-Disposition: attachment; filename=\"$file\"\n\n";
	$zip->writeToFileHandle(*STDOUT);
}

sub restore{
	eval "use Archive::Zip;";
	return if($@);

	my $boundary = <STDIN>;
	my $file = <STDIN>; my $tmp = $file.<STDIN>.<STDIN>;
	$file =~ s#^.*?filename="(.*?)".*$#$1#s; $file =~ s#^.*[/\\]##;
	my $length = $CONTENT_LENGTH - length($tmp) - 2 *
		length($boundary) - 4;
	read STDIN, my $content, $length;
	my ($fh, $name) = Archive::Zip::tempFile(".");
	print $fh $content;
	my $zip = Archive::Zip->new();
	$zip->readFromFileHandle($fh);
	(my $u_cgi = $U_CGI) =~ s#^(?:/~[^/]+)?/##;
	$zip->removeMember($u_cgi);
	$zip->removeMember($PW);
	foreach($zip->memberNames()){
		$zip->removeMember($_) if(-f $_ && !-w $_);
		if($HOSTING eq "awardspace" &&
			(m#\.html$# || m#/#)){
			my $member = $zip->memberNamed($_);
			$member->unixFileAttributes(0755);
		}
	}
	$zip->extractTree();
	close $fh;
	unlink $name;

	if($HOSTING eq "awardspace"){
		foreach($zip->memberNames()){
			chmod 0755, $_ if(m#\.html$# || m#/#);
		}
	}
}

sub verify_input{
	return 1;
}

sub comment_form{
	my ($page, $comment, $direction, $rows, $cols) = @_;
	$page = $PAGE if($page eq "");
	$comment = "comment" if($comment eq "");
	$direction = "down" if($direction eq "");
	$rows = "5" if($rows eq "");
	$cols = "80" if($cols eq "");

	return <<EOT;
<form class="comment_input" action="$U_CGI?comment=$comment" method="post">
<div>
<input type="hidden" id="page" name="page" value="$page" />
<input type="hidden" id="direction" name="direction" value="$direction" />
<textarea accesskey="c" id="text" name="text" rows="$rows" cols="$cols"></textarea>
<input type="submit" value="write" />
</div>
</form>
EOT
}

sub goto_form{
	return <<EOT;
<form class="goto_input" action="$U_CGI" method="get">
<div>
<input accesskey="g" id="goto" name="goto" />
<input type="submit" value="goto" />
</div>
</form>
EOT
}

sub search_form{
	return <<EOT;
<form class="search_input" action="$U_CGI" method="get">
<div>
<input accesskey="s" id="search" name="search" />
<input type="submit" value="search" />
<input type="checkbox" id="simple" name="simple" value="1" checked="checked" /> simple
<input type="checkbox" id="link" name="link" value="1" /> link
<input type="checkbox" id="icase" name="icase" value="1" checked="checked" /> ignore case
<input type="checkbox" id="title" name="title" value="1" /> display title
<input type="checkbox" id="nomatch" name="nomatch" value="1" /> no match
</div>
</form>
EOT
}

################################################################################
# text parser
sub line{
	local $_ = shift;
	s/[\r\n]//g;

	if($wiki && !$pre){
		return if(m/^#\$[^#\$]/);
		s/\[ (.+?) \]/['''' $1 '''']/g;
	}

	if($block ne ""){
		if($_ eq "#\$}"){
			my $i = "$block}";
			undef $block;
			$i =~ s/^(#\$[{}]_*)_$/$1/mg;
			eval $i;
			return;
		}
		$block .= "$_\n";
		return;
	}
	if(m/^(?:##|[^#]|$)/){
		unless($pre){
			for(my $i=$reg_i0; $i<$reg_i; $i++){
				eval "s\x00$reg[$i]\x00$reg_sub[$i]\x00g;";
			}
			if(m/\n/){
				local $reg_i0 = $reg_i;
				my @i = split /\n/, "$_\n\x01";
				$#i--;
				foreach(@i){
					if($pre || m/^#/){
						s/\x01/&/g;
						s/\x02/</g;
						s/\x03/>/g;
					}
					$line->($_);
				}
				return;
			}
		}
	}
	if(m/^(?:[^|#]|$)/){
		if($table){
			$text .= "</table>\n<$p>";
			$table = 0;
		}
	}
	if(m/^---+$/ && (!$pre || $pre == length)){
		if($pre){
			$text .= "</pre>\n<$p>";
			$pre = 0;
		}else{
			$text .= "</$p>\n<pre>\n";
			$pre = length;
		}
		return;
	}
	if($pre){
		s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g; s/\x00//g;
		$text .= "$_\n";
		return;
	}
	if(m/^#\$include (.*)$/){
		local *FH;
		if(open FH, $1){
			$line->($_) while(<FH>);
			close FH;
		}
		return;
	}
	if(m/^#\$shell (.*)$/){
		$_ = $1;
		s/\x01/&/g; s/\x02/</g; s/\x03/>/g; s/\x00//g;
		$text .= `$_`;
		return;
	}
	if(m/^#\$((?:sub |\{).*)$/){
		if("}" eq substr $1, -1){
			eval $1;
			return;
		}
		$block = "$1\n";
		return;
	}
	if(m/^#\$[^#\$]/){
		return;
	}
	if(m/^#\$\$(.*)$/){
		if($1 eq ""){
			$reg_i = 0;
			$#reg = $#reg_sub = -1;
		}else{
			for(my $i=0; $i<$reg_i; $i++){
				if($reg[$i] eq $1){
					$reg_i--;
					for(; $i<$reg_i; $i++){
						$reg[$i] = $reg[$i+1];
						$reg_sub[$i] = $reg_sub[$i+1];
					}
					$#reg = $#reg_sub = $reg_i-1;
					last;
				}
			}
		}
		return;
	}
	if(m/^#([^#]+)#(.*)$/){
		# Don't allow delimiters
                return if(index($1, "\x00") != -1 || index($2, "\x00") != -1);

                # Don't allow code embedding in a wiki page
                return if($wiki && (index($1, '(?{') != -1 ||
                                    index($1, '(??{') != -1 ||
                                    index($2, '@{[') != -1));

		my $i;
		for($i=0; $i<$reg_i; $i++){
			last if($reg[$i] eq $1);
		}
		$reg[$reg_i++] = $1 if($i == $reg_i);
		$_ = $2;
		s/&/\x01/g; s/</\x02/g; s/>/\x03/g; s/\$/\\\$/g;
		$reg_sub[$i] = $_;
		return;
	}
	if(m/^##(.*)$/){
		$_ = $1;
		s/\x01/&/g; s/\x02/</g; s/\x03/>/g; s/\x00//g;
		$text .= "$_\n";
		return;
	}
	if(m/^#/){
		return;
	}
	if(m/^TOC$/){
		$text .= "\x05\n";
		return;
	}
	if(m/^___+$/){
		$text .= "</$p>\n<hr />\n<$p>";
		return;
	}
	if(m/^$/){
		$li_p[$li_i-1] = 1 if($li_i > 0);
		$text .= "</$p>\n<$p>";
		return;
	}
	s/&/&amp;/g; s/</\x04&lt;/g; s/>/\x04&gt;/g;
	s/\x01/&/g; s/\x02/</g; s/\x03/>/g;
	s#''(.*?)''#<i>$1</i>#g;
	s#``(.*?)``#<b>$1</b>#g;
	s#""(.*?)""#<tt>$1</tt>#g;
	s#\[([^\]< \t]+)\.(\#[^\] \t]*)? ([^\]]*[^\] \t])\]#<a href="$1.html$2">$3</a>#g;
	s#\[([^\]< \t]+)\.(\#[^\] \t]*)?\]#<a href="$1.html$2">$1$2</a>#g;
	s#\[([^\]< \t]+) ([^\]]*[^\] \t])\]#<a href="\x04$1">$2</a>#g;
	s#\[([^\]< \t]+)\]#<a href="\x04$1">$1</a>#g;
	s#(<[^>]*)((?:$protocol)[^>]*>)#$1\x04$2#ogi;
	s#(<a [^>]*>[^<]*)((?:$protocol)[^<]*</a>)#$1\x04$2#ogi;
	s#(^|[^a-zA-Z0-9\x04])((?:$protocol)[$protocol_char]+&(amp;)?[a-z]+;)(?=(?:[ \t]|$))#$1<a href="\x04$2">\x04$2</a>#ogi;
	s#(^|[^a-zA-Z0-9\x04])((?:$protocol)[$protocol_char]+)(?=[$protocol_punct](?:[ \t]|$))#$1<a href="\x04$2">\x04$2</a>#ogi;
	s#(^|[^a-zA-Z0-9\x04])((?:$protocol)[$protocol_char]+)#$1<a href="\x04$2">\x04$2</a>#ogi;
	s#<a href="\x04([^"]+\.(?:$image_ext))">([^<]+)</a>#<img src="$1" alt="$2" title="$2" />#ogi;
	s/\x04//g;
	if(m/^( *)([-+]|\{.*?\}) /){
		my $i = length($1)+1;
		my $j;
		my $tag = substr $2, 0, 1;
		if($tag eq "-"){
			$tag = "ul";
		}elsif($tag eq "+"){
			$tag = "ol";
		}else{
			$tag = "dl";
		}
		if($i > $li_i){
			for(; $li_i<$i-1; $li_i++){
				$j = $li_i + 1;
				$text .= "</p:$li_i>\n<$tag>\n";
				if($tag eq "dl"){
					$text .= "<dd><p:$j>";
				}else{
					$text .= "<li><p:$j>";
				}
				$li[$li_i] = $tag;
			}
			$text .= "</p:$li_i>\n<$tag>\n";
			$li[$li_i++] = $tag;
		}elsif($i < $li_i){
			while(--$li_i>=$i){
				$j = $li_i + 1;
				if($li[$li_i] eq "dl"){
					$text .= "</p:$j></dd>\n";
				}else{
					$text .= "</p:$j></li>\n";
				}
				$text .= "</$li[$li_i]>\n<p:$li_i>";
				if($li_p[$li_i]){
					$text =~ s#\n?(</?p):$j(>)#$1$2#g;
					$li_p[$li_i] = 0;
				}else{
					$text =~ s#\n?</?p:$j>##g;
				}
			}
			if($li[$li_i] eq $tag){
				$j = $li_i + 1;
				if($li[$li_i] eq "dl"){
					$text .= "</p:$j></dd>\n";
				}else{
					$text .= "</p:$j></li>\n";
				}
				$li_i++;
			}else{
				for(; $li_i>=0 && $li[$li_i] ne $tag; $li_i--){
					$j = $li_i + 1;
					if($li[$li_i] eq "dl"){
						$text .= "</p:$j></dd>\n";
					}else{
						$text .= "</p:$j></li>\n";
					}
					$text .= "</$li[$li_i]>\n<p:$li_i>";
					if($li_p[$li_i]){
						$text =~ s#\n?(</?p):$j(>)#$1$2#g;
						$li_p[$li_i] = 0;
					}else{
						$text =~ s#\n?</?p:$j>##g;
					}
				}
				while(++$li_i<$i){
					$text .= "</p:$li_i>\n<$tag>";
					$li[$li_i] = $tag;
				}
			}
		}elsif($li[$li_i-1] ne $tag){
			if($li[$li_i-1] eq "dl"){
				$text .= "</p:$li_i></dd>\n";
			}else{
				$text .= "</p:$li_i></li>\n";
			}
			$text .= "</$li[$li_i-1]>\n<$tag>\n";
			if($li_p[$li_i-1]){
				$text =~ s#\n?(</?p):$li_i(>)#$1$2#g;
				$li_p[$li_i-1] = 0;
			}else{
				$text =~ s#\n?</?p:$li_i>##g;
			}
			$li[$li_i-1] = $tag;
		}elsif($tag eq "dl"){
			$text .= "</p:$li_i></dd>\n";
		}else{
			$text .= "</p:$li_i></li>\n";
		}

		$p = "p:$li_i";
		if($tag eq "dl"){
			s# *\{(.*?)\} (.*)#<dt>$1</dt>\n<dd><$p>$2#;
		}else{
			s# *[-+] (.*)#<li><$p>$1#;
		}
	}
	if(m/^( *)(?:[-+]|\{\})$/){
		my $j;
		for(my $i=length($1); --$li_i>=$i; ){
			$j = $li_i + 1;
			if($li[$li_i] eq "dl"){
				$text .= "</p:$j></dd>\n";
			}else{
				$text .= "</p:$j></li>\n";
			}
			$text .= "</$li[$li_i]>\n<p:$li_i>";
			if($li_p[$li_i]){
				$text =~ s#\n?(</?p):$j(>)#$1$2#g;
				$li_p[$li_i] = 0;
			}else{
				$text =~ s#\n?</?p:$j>##g;
			}
		}
		if(++$li_i){
			$p = "p:$li_i";
		}else{
			$p = "p";
		}
		return;
	}
	if(m/^\|(?:[0-9]*,?[0-9]*)? /){
		my $i = 0;
		if(" |" eq substr $_, -2){
			$_ = substr $_, 0, length($_)-2;
			$i = 1;
		}
		unless($table){
			if($i){
				$text .= "</$p>\n<table class=\"uniqki_table\">\n";
			}else{
				$text .= "</$p>\n<table>\n";
			}
			$table = 1;
		}
		s#^\|([0-9]+),([0-9]+)   #<tr><td rowspan="$2" colspan="$1" align="right">#;
		s#^\|([0-9]+),([0-9]+)  #<tr><td rowspan="$2" colspan="$1" align="center">#;
		s#^\|([0-9]+),([0-9]+) #<tr><td rowspan="$2" colspan="$1" align="left">#;
		s#^\|,([0-9]+)   #<tr><td rowspan="$1" align="right">#;
		s#^\|,([0-9]+)  #<tr><td rowspan="$1" align="center">#;
		s#^\|,([0-9]+) #<tr><td rowspan="$1" align="left">#;
		s#^\|([0-9]+)   #<tr><td colspan="$1" align="right">#;
		s#^\|([0-9]+)  #<tr><td colspan="$1" align="center">#;
		s#^\|([0-9]+) #<tr><td colspan="$1" align="left">#;
		s#^\|   #<tr><td align="right">#;
		s#^\|  #<tr><td align="center">#;
		s#^\| #<tr><td align="left">#;
		s# \|([0-9]+),([0-9]+)   #</td><td rowspan="$2" colspan="$1" align="right">#g;
		s# \|([0-9]+),([0-9]+)  #</td><td rowspan="$2" colspan="$1" align="center">#g;
		s# \|([0-9]+),([0-9]+) #</td><td rowspan="$2" colspan="$1" align="left">#g;
		s# \|,([0-9]+)   #</td><td rowspan="$1" align="right">#g;
		s# \|,([0-9]+)  #</td><td rowspan="$1" align="center">#g;
		s# \|,([0-9]+) #</td><td rowspan="$1" align="left">#g;
		s# \|([0-9]+)   #</td><td colspan="$1" align="right">#g;
		s# \|([0-9]+)  #</td><td colspan="$1" align="center">#g;
		s# \|([0-9]+) #</td><td colspan="$1" align="left">#g;
		s# \|   #</td><td align="right">#g;
		s# \|  #</td><td align="center">#g;
		s# \| #</td><td align="left">#g;
		$_ .= "</td></tr>";
	}
	if(m/^( *=)!? /){
		my $i = length($1);
		my $inc_toc = (m/^ *= /?1:0);
		s#^ *=!? ##;
		s/\x00//g;
		s#\[ ([^\]]*[^\] \t][^\]]*) \]#$1#eeg;
		(my $t = $_) =~ s/<[^>]*>//g;
		$t =~ s/^ *//; $t =~ s/ *$//;
		my $id = $t;
		$id =~ s/&amp;/&/g; $id =~ s/&lt;/</g; $id =~ s/&gt;/>/g;
		#$id =~ s/([\t`~!@#\$%^&*_+=\\|;'",\/?()\[\]{}<>])/sprintf(".%02X", ord($1))/eg;
		#$id =~ y/ /_/;
		$id =~ s/[ \t`~!@#\$%^&*_+=\\|;'",\/?()\[\]{}<>]+/_/g;
		$id =~ s/_+$//;
		my $j = $h_i{$id}++;
		for(my $i=0; $i<$j; $i++){
			$id .= "_";
		}
		$id = "_$id";
		if($i > 6){
			$text .= qq(</$p>\n<div id="$id" class="uniqki_h7">$_</div>\n<$p>\n);
		}else{
			$text .= qq(</$p>\n<h$i id="$id">$_</h$i>\n<$p>\n);
			if($TITLE eq "" && $i == 1){
				$TITLE = $_;
				$TITLE =~ s/<[^>]*>//g;
			}
		}
		return unless($inc_toc);

		if($i > $h_prev){
			if($h_prev){
				$toc .= "<dd><dl>" while($h_prev++<$i);
				$h_prev--;
			}else{
				$toc .= "<dl class=\"uniqki_toc\">";
				$toc .= "\n" if($i > 1);
				$toc .= "<dd><dl>" while(++$h_prev<$i);
				$h_top = $i;
			}
			$toc .= "\n";
		}elsif($i < $h_prev){
			$toc .= "</dl></dd>" while(--$h_prev>=$i);
			$toc .= "\n";
			$h_prev++;
			$h_top = $i if($i < $h_top);
		}
		$toc .= "<dt><a href=\"#$id\">$t</a></dt>\n";
		return;
	}
	s/\x00//g;
	s#\[ ([^\]]*[^\] \t][^\]]*) \]#$1#eeg;
	$text .= "$_\n";
}

sub end{
	if($pre){
		$text .= "</pre>\n<$p>";
		$pre = 0;
	}
	if($table){
		$text .= "</table>\n<$p>";
		$table = 0;
	}
	if($li_i){
		my $i;
		while(--$li_i>=0){
			$i = $li_i + 1;
			if($li[$li_i] eq "dl"){
				$text .= "</p:$i></dd>\n";
			}else{
				$text .= "</p:$i></li>\n";
			}
			$text .= "</$li[$li_i]>\n<p:$li_i>";
			if($li_p[$li_i]){
				$text =~ s#\n?(</?p):$i(>)#$1$2#g;
				$li_p[$li_i] = 0;
			}else{
				$text =~ s#\n?</?p:$i>##g;
			}
		}
		$li_i = 0;
	}
	$text =~ s#\n?(</?p):0(>)#$1$2#g;
	if($h_prev){
		my $i = $h_prev;
		$toc .= "</dl></dd>" while(--$h_prev>0);
		$toc .= ($i>1?"\n":"")."</dl>\n";
		for(my $i=0; $i<$h_top-1; $i++){
			$toc =~ s#^(.*\n)<dd><dl>#\1#;
			$toc =~ s#</dl></dd>(\n.*)$#\1#;
		}
		$toc =~ s#\n+#\n#g;
		$toc = "</p>\n$toc<p>";
		$h_prev = 0;
	}
	$text =~ s/\x05\n?/$toc/g;
}

sub text{
	my @i = split /\n/, "$_[0]\x01"; $#i--;
	$line->($_) foreach(@i);
}

sub pre_text{
	$protocol = 'https?://|ftp://|news://|telnet://|gopher://|wais://|mailto:|file://';
	$protocol_char = 'a-zA-Z0-9:@/~%.,_$?=&;#+-';
	$protocol_punct = '.,;:!?';
	$image_ext = 'png|gif|jpg|jpeg';

	$text = "<p>";
	$block = "";
	$reg_i0 = 0;
	$reg_i = 0;
	@reg = ();
	@reg_sub = ();
	$toc = "";
	%h_i = ();
	$h_top = 0;
	$h_prev = 0;
	$pre = 0;
	$table = 0;
	$li_i = 0;
	@li = ();
	@li_p = ();
	$p = "p";
}

sub post_text{
	$text .= "</p>\n";
	$text =~ s#<(i|b|tt)>([ \t\n]*)</\1>#$2#g;
	$text =~ s#<p>[ \t\n]*</p>##g;
	$text =~ s#(<p>)\n+#$1#g; $text =~ s#\n+(</p>)#$1#g;
}

sub uniqki{
	my $txt = shift;

	local *UNIQKI_FH;
	return unless(open UNIQKI_FH, "<", $txt);
	$wiki = (<UNIQKI_FH> eq "#!wiki\n" ? 1 : 0);
	close UNIQKI_FH;

	local ($protocol, $protocol_char, $protocol_punct, $image_ext, $text,
		$block, $reg_i0, $reg_i, @reg, @reg_sub, $toc, %h_i, $h_top,
		$h_prev, $pre, $table, $li_i, @li, @li_p, $p);
	my ($header, $footer);

	unless($wiki){
		($header, $footer) = ($HEADER, $FOOTER);
	}else{
		($header, $footer) = ($WIKI_HEADER, $WIKI_FOOTER);
	}

	$pre_text = \&pre_text unless(defined($pre_text));
	$line = \&line unless(defined($line));
	$end = \&end unless(defined($end));
	$post_text = \&post_text unless(defined($post_text));

	$pre_text->();
	foreach my $file ($header, $txt, $footer){
		# "<" is required for the in-memory file.
		next if($file eq "" || !open UNIQKI_FH, "<", $file);
		$line->($_) while(<UNIQKI_FH>);
		close UNIQKI_FH;
	}
	$end->();
	$post_text->();

	return $text;
}

################################################################################
# user actions
if($QUERY_STRING =~ m/^goto=(.+)$/){
#-------------------------------------------------------------------------------
# u.cgi?goto=PAGE		go to or create PAGE using a form (admin only)
	$_ = $1;
	s#/.*$##; s#\.html$##;
	y/+/ /;

	print "Location: $HTTP_BASE$SCRIPT_NAME/$_\n\n";
	exit;
}elsif($QUERY_STRING eq "refresh" && $PAGE ne ""){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?refresh		refresh
	make_html($PAGE) if(-f "$PAGE.txt");
	print "Location: $DOC_BASE/$PAGE.html\n\n";
	exit;
}elsif($QUERY_STRING =~ m/^diff(?:=(-?[0-9]+))?$/){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?diff		diff current and previous version
# u.cgi/PAGE?diff=([0-9]+)	diff current and \1 version
# u.cgi/PAGE?diff=-([0-9]+)	diff current and current-\1 version
	exit unless(-f "$PAGE.txt");

	my $current_version = get_version($PAGE);
	my $version = ($1 > 0 ? $1 : $current_version + ($1 < 0 ? $1 : -1));
	if($version >= $current_version || $version <= 0){
		print "Content-Type: text/plain\n\n".
			"Current version: $current_version\n";
		exit;
	}

	local *FH;
	open FH, "$PAGE.txt"; local $/ = undef;
	my $current_text = <FH>;
	close FH;

	my $text = $current_text;
	open FH, "$PAGE.txt.v"; local $/ = "\x01\n";
	while(<FH>){
		m/^(.*?)\n(.*)\x01\n$/s;
		$text = patch($text, $2);
		last if($version == $1);
	}
	close FH;

	print "Content-Type: text/html\n\n";
	do "$TPL/header.tpl";
	print qq(<div id="diff">\n);

	my @line0 = split /\n/, "$text\x01";
	my @line1 = split /\n/, "$current_text\x01";
	$#line0--; $#line1--;
	my ($s, @delta) = lcs(\@line0, \@line1);
	my $m = $s;
	my $n;
	for($n=0; $n<$s; $n++){
		$_ = $line1[$n];
		s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
		print qq(<div class="diff_unchanged">= $_</div>\n);
	}
	eval "use Encode;";
	my $encode = ($@ ? 0 : 1);
	for(my $i=0; $i<=$#delta; $i++,$m++,$n++){
		my ($x, $y) = split /,/, $delta[$i];
		if($x > $m && $y > $n){
			for(; $m<$x&&$n<$y; $m++,$n++){
				print qq(<div class="diff_modified">* );
				my $l0 = "$line0[$m]\x01";
				my $l1 = "$line1[$n]\x01";
				if($encode){
					$l0 = Encode::decode($CHARSET, $l0);
					$l1 = Encode::decode($CHARSET, $l1);
				}
				my @l0 = split //, $l0;
				my @l1 = split //, $l1;
				my ($is, @idelta) = lcs(\@l0, \@l1);
				$#l0--; $#l1--;
				my $im = $is;
				my $in;
				for($in=0; $in<$is; $in++){
					$_ = $l1[$in];
					s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
					print;
				}
				for(my $ii=0; $ii<=$#idelta; $ii++,$im++,$in++){
					my ($ix, $iy) = split /,/, $idelta[$ii];
					if($ix > $im){
						print qq(<span class="diff_modified_deleted">);
						$_ = "";
						for(; $im<$ix; $im++){
							$_ .= $l0[$im];
						}
						s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
						print qq($_</span>);
					}
					if($iy > $in){
						print qq(<span class="diff_modified_added">);
						$_ = "";
						for(; $in<$iy; $in++){
							$_ .= $l1[$in];
						}
						s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
						print qq($_</span>);
					}
					if($in <= $#l1){
						$_ = $l1[$in];
						s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
						print;
					}
				}
				for(; $in<=$#l1; $in++){
					$_ = $l1[$in];
					s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
					print;
				}
				print qq(</div>\n);
			}
		}
		if($x > $m){
			for(; $m<$x; $m++){
				$_ = $line0[$m];
				s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
				print qq(<div class="diff_deleted">- $_</div>\n);
			}
		}
		if($y > $n){
			for(; $n<$y; $n++){
				$_ = $line1[$n];
				s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
				print qq(<div class="diff_added">+ $_</div>\n);
			}
		}
		if($n <= $#line1){
			$_ = $line1[$n];
			s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
			print qq(<div class="diff_unchanged">= $_</div>\n);
		}
	}
	for(; $n<=$#line1; $n++){
		$_ = $line1[$n];
		s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
		print qq(<div class="diff_unchanged">= $_</div>\n);
	}

	print qq(</div>\n);
	do "$TPL/footer.tpl";
	exit;
}elsif($QUERY_STRING =~ m/^ls(?:[&=].+)?$/){
#-------------------------------------------------------------------------------
# u.cgi?ls			list all pages in alphabetical order
# u.cgi?ls=az			list all pages in alphabetical order
# u.cgi?ls=za			list all pages in reverse order
# u.cgi?ls=taz			list all pages in alphabetical order of titles
# u.cgi?ls=tza			list all pages in reverse order of titles
# u.cgi?ls=rc			recent changes
# u.cgi?ls=oc			old changes
# u.cgi?ls&n=([0-9]+)		list only \1 pages in alphabetical order
# u.cgi?ls&title=1		display page titles instead of page names
# u.cgi?ls&glob=GLOB		list all GLOB pages in alphabetical order
	my %var = get_var();
	my $glob = ($var{glob} eq "" ? "*" : $var{glob});

	$PAGE = ($var{ls} eq "rc" ? "recent changes" :
		($var{ls} eq "oc" ? "old changes" :
		($var{ls} eq "za" ? "all pages (Z to A)" : "all pages"))).
		" matching $glob";

	print "Content-Type: text/html\n\n";
	do "$TPL/header.tpl";
	print qq(<div id="ls">\n<h1>$PAGE</h1>\n);

	my $tls = ($var{ls} eq "taz" || $var{ls} eq "tza" ? 1 : 0);
	my $roc = ($var{ls} eq "rc" || $var{ls} eq "oc" ? 1 : 0);
	my $n = ($var{n} eq "" ? 0 : $var{n});
	my $i = 0;
	my @list;
	foreach(<$glob.txt>){
		next if(substr($_,0,1) eq "." || index($_,"/") >= 0 || !-f $_);
		s/\.txt$//;
		my $page = $_;
		my $title = $page;
		my @t = localtime((stat "$page.txt")[9]);
		my $time = sprintf "%d-%02d-%02d %02d:%02d:%02d",
			$t[5]+1900, $t[4]+1, $t[3], $t[2], $t[1], $t[0];
		if($tls || $var{title} eq "1"){
			local *FH;
			if(open FH, "$page.html"){
				local $/ = undef;
				my $text = <FH>;
				close FH;
				$text =~ s#<(script|style).*?</\1>##sgi;
				if($text =~ m#<h1[^>]*>(.+?)</h1>#si){
					$_ = $1;
					s/<[^>]*>//g; s/[<>]//g;
					$title = $_;
				}
			}
		}
		if($roc){
			if($var{title} eq "1"){
				$list[$i++] = "$time\x00$title\x00$page";
			}else{
				$list[$i++] = "$time\x00$page\x00$title";
			}
		}elsif($tls){
			$list[$i++] = "$title\x00$page\x00$time";
		}else{
			$list[$i++] = "$page\x00$title\x00$time";
		}
	}
	@list = sort @list;
	@list = reverse @list
		if($var{ls} eq "rc" || $var{ls} eq "za" || $var{ls} eq "tza");
	$n = $i if($n > $i || !$n);
	$i = 0;
	foreach(@list){
		my ($page, $title, $time);
		m/^(.*)\x00(.*)\x00(.*)$/;
		if($roc){
			if($var{title} eq "1"){
				$page = $3;
				$title = $2;
			}else{
				$page = $2;
				$title = $3;
			}
			$time = $1;
		}else{
			if($tls){
				$page = $2;
				$title = ($var{title} eq "1" ? $1 : $page);
			}else{
				$page = $1;
				$title = $2;
			}
			$time = $3;
		}
		print qq(<div><a href="$DOC_BASE/$page.html">$title</a> <span class="ls_time">$time</span></div>\n);
		last if(++$i == $n);
	}

	print qq(</div>\n);
	do "$TPL/footer.tpl";
	exit;
}elsif($QUERY_STRING =~ m/^rss(?:[&=].+)?$/){
#-------------------------------------------------------------------------------
# u.cgi?rss			RSS for recent 10 pages
# u.cgi?rss=([0-9]+)		RSS for recent \1 pages
# u.cgi?rss&glob=GLOB		RSS for recent 10 GLOB pages
	my %var = get_var();
	my $glob = ($var{glob} eq "" ? "*" : $var{glob});

	my $t = time;
	my @g = gmtime $t;
	my @l = localtime $t;
	my $g = $g[5]*31536000+$g[4]*2592000+$g[3]*86400+$g[2]*3600+$g[1]*60+
		$g[0];
	my $l = $l[5]*31536000+$l[4]*2592000+$l[3]*86400+$l[2]*3600+$l[1]*60+
		$l[0];
	my $i = 0;
	my @list;
	foreach(<$glob.txt>){
		next if(substr($_,0,1) eq "." || index($_,"/") >= 0 || !-f $_);
		s/\.txt$//;
		my $t = (stat "$_.txt")[9];
		my @t = split / +/, scalar gmtime($t);
		$list[$i++] = sprintf "%d %s, %02d %s %d %s GMT %s",
			$t, $t[0], $t[2], $t[1], $t[4], $t[3], $_;
	}
	my $n = 10;
	$n = $1 if($var{rss} =~ m/^([0-9]+)$/);
	$n = $i if($n > $i || !$n);
	print <<EOT;
Content-Type: text/xml

<?xml version="1.0" encoding="$CHARSET"?>
<rss version="2.0">
<channel>
<title>$RSS_TITLE</title>
<link>$DOC_BASE</link>
<description>$RSS_DESCRIPTION</description>
EOT
	undef $/;
	$i = 0;
	foreach(reverse sort @list){
		my ($time, $page) = m/^[0-9]+ (.+? GMT) (.*)$/;
		local *FH;
		open FH, "$page.html";
		my $txt = <FH>;
		close FH;
		$txt =~ s/\r//g;
		$txt =~ s#^.*<body[^>]*>##si;
		$txt =~ s#</body>.*$##si;
		my $has_more = ($txt =~ s'<!-- #more -->.*$''s);
		$txt =~ s'<!-- # -->.*''g;
		my $title;
		if($txt =~ m#<h1[^>]*>(.+?)</h1>(.*)$#si){
			$title = $1;
			$txt = $2;
			$title =~ s/<[^>]*>//g;
			$title =~ s/&[^ ]*;/ /g;
			$title =~ s/[ \t\n]+/ /g;
			$title =~ s/^ //;
			$title =~ s/ $//;
		}else{
			$title = $page;
		}
		$txt =~ s#<(script|style).*?</\1>##sgi;
		$txt =~ s/<[^>]*>//g;
		$txt =~ s/&[^ ]*;/ /g;
		$txt =~ s/[ \t\n]+/ /g;
		$txt =~ s/^ //;
		$txt =~ s/ $//;
		if($txt =~ m/^((?:[^ ]+ ){20})/){
			$txt = "$1...";
		}elsif($has_more){
			$txt .= " ...";
		}
		print <<EOT;
<item>
<title>$title</title>
<link>$DOC_BASE/$page.html</link>
<description>$txt</description>
<pubDate>$time</pubDate>
</item>
EOT
		last if(++$i == $n);
	}
	print <<EOT;
</channel>
</rss>
EOT
	exit;
}elsif($QUERY_STRING =~ m/^search=.+$/){
#-------------------------------------------------------------------------------
# u.cgi?search=(.*)		search using regular expressions
# u.cgi?search=(.*)&simple=1	search using space separated words
# u.cgi?search=(.*)&icase=1	case-insensitive search
# u.cgi?search=(.*)&link=1	search for pages which link to \1 page
# u.cgi?search=(.*)&title=1	display page titles instead of page names
# u.cgi?search=(.*)&nomatch=1	don't display matches
# u.cgi?search=(.*)&glob=GLOB	search GLOB pages
	my %var = get_var();
	my $glob = ($var{glob} eq "" ? "*" : $var{glob});

	$_ = $var{search};
	if($var{simple} eq "1"){
		s/[<>\r]//g; s/[\t\n]/ /g; s/ +/ /g; s/^ //; s/ $//;
		$_ = quotemeta;
		s/\\ />/g;
		s/\\"([^"]*)\\"/\x01$1\x02/g;
		while(s/\x01([^\x02]*)>([^\x02]*)\x02/\x01$1 +$2\x02/g){}
		s/[\x01\x02]//g; s/\\"//g;
	}
	my @search = split />/;

	$_ = $var{search};
	s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
	my $title = $_;
	$PAGE = "search $glob for $title";

	print "Content-Type: text/html\n\n";
	do "$TPL/header.tpl";
	print qq(<div id="search">\n<h1>$PAGE</h1>\n);

	foreach(<$glob.html>){
		next if(substr($_,0,1) eq "." || index($_,"/") >= 0 || !-f $_);
		s/\.html$//;
		my $page = $title = $_;
		local *FH;
		next unless(open FH, "$page.html");
		local $/ = undef;
		my $text = <FH>;
		close FH;

		$text =~ s#<(script|style).*?</\1>##sgi;
		if($var{title} eq "1" && $text =~ m#<h1[^>]*>(.+?)</h1>#si){
			$_ = $1;
			s/<[^>]*>//g; s/[<>]//g;
			$title = $_;
		}
		my ($i, $line, $search, $found, @found, @result);
		if($var{link} eq "1"){
			foreach(split /\n/, $text){
				$line = $_;
				$found = "";
				for($i=0; $i<=$#search; $i++){
					$search = $search[$i];
					$search = "(?i)$search"
						if($var{icase} eq "1");
					if($line =~ m#<[aA][ \t][^>]*(?:href|HREF)[ \t]*=[ \t]*"(?:$search)"[^>]*>[^<]+</[aA]>#){
						$found = $line if($found eq "");
						$found =~ s#<[aA][ \t][^>]*(?:href|HREF)[ \t]*=[ \t]*"(?:$search)"[^>]*>([^<]+)</[aA]>#\x01$1\x02#g;
						$found[$i] = 1;
					}
				}
				if($found ne ""){
					$_ = $found;
					s/<[^>]*>//g; s/<.*$//; s/^.*>//;
					push @result, $_;
				}
			}
		}else{
			$text =~ s/<[^>]*>//g; $text =~ s/[<>]//g;
			foreach(split /\n/, $text){
				$line = $_;
				$found = "";
				for($i=0; $i<=$#search; $i++){
					$search = $search[$i];
					$search = "(?i)$search"
						if($var{icase} eq "1");
					if($line =~ m#$search#){
						$found = $line if($found eq "");
						$found =~ s#($search)#\x01$1\x02#g;
						$found[$i] = 1;
					}
				}
				push @result, $found if($found ne "");
			}
		}
		for($i=0; $i<=$#search; $i++){
			last if(!$found[$i]);
		}
		next if($i <= $#search);
		if($var{nomatch} eq "1"){
			print qq(<div><a href="$DOC_BASE/$page.html">$title</a></div>\n);
			next;
		}
		foreach(@result){
			s#\x01#<span class="search_highlight">#g;
			s#\x02#</span>#g;
			print qq(<div><a href="$DOC_BASE/$page.html">$title</a>: $_</div>\n);
		}
	}

	print qq(</div>\n);
	do "$TPL/footer.tpl";
	exit;
}elsif($QUERY_STRING =~ m/^comment(?:[&=].+)?$/){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?comment		generate the comment form for PAGE
#				(id=comment, downward, rows=80, cols=5)
# u.cgi/PAGE?comment&direction=(up|down) generate the comment form for PAGE
#				(id=comment, (up|down)ward, rows=80, cols=5)
# u.cgi/PAGE?comment&rows=([0-9]+) generate the comment form for PAGE
#				(id=comment, downward, rows=\1, cols=5)
# u.cgi/PAGE?comment&cols=([0-9]+) generate the comment form for PAGE
#				(id=comment, downward, rows=80, cols=\1)
# u.cgi/PAGE?comment=COMMENT	generate the comment form for PAGE
#				(id=COMMENT, downward, rows=80, cols=5)
# u.cgi?comment&page=PAGE	generate the comment form for PAGE
#				(id=comment, downward, rows=80, cols=5)
	my %var = get_var();
	if($REQUEST_METHOD eq "GET"){
		return if($PAGE eq "" && $var{page} eq "");
		print "Content-Type: text/html\n\n";
		do "$TPL/header.tpl";
		print comment_form($var{page}, $var{comment}, $var{direction},
			$var{rows}, $var{cols});
		do "$TPL/footer.tpl";
		exit;
	}
	exit unless(verify_input("comment", \%var));

	$PAGE = $var{page};
	$var{text} =~ s/$/ /mg;

	mylock("$PAGE.txt");
	my $TEXT = "";
	my $date = scalar localtime;
	my $added = 0;
	local *FH;
	if(open FH, "$PAGE.txt"){
		while(<FH>){
			if(m/^#$var{comment}$/){
				if($var{direction} eq "up"){
					$TEXT .= "$_# $date\n$var{text}\n\n";
				}else{
					$TEXT .= "# $date\n$var{text}\n\n$_";
				}
				$added = 1;
			}else{
				$TEXT .= $_;
			}
		}
		close FH;
	}
	unless($added){
		myunlock("$PAGE.txt");
		exit;
	}
	save($PAGE, $TEXT);
	myunlock("$PAGE.txt");

	make_html($PAGE);

	print "Location: $DOC_BASE/$PAGE.html\n\n";
	exit;
}elsif($FILE ne "" && $QUERY_STRING ne "delete"){
#-------------------------------------------------------------------------------
# u.cgi/PAGE/FILE		show/download PAGE/FILE
	print "Location: $DOC_BASE/$PAGE/$FILE?$QUERY_STRING\n\n";
	exit;
}elsif($QUERY_STRING =~ m/^wiki/){
#-------------------------------------------------------------------------------
# uniqkiwiki
if($REQUEST_METHOD eq "GET"){
	local *FH;
	local $TEXT;
	if($QUERY_STRING eq "wikiedit"){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?wikiedit		edit wiki PAGE
		$TEXT = "";
		if(-f "$PAGE.txt"){
			open FH, "$PAGE.txt";
			if(<FH> ne "#!wiki\n"){
				close FH;
				exit;
			}
			while(<FH>){
				s/&/&amp;/g; s/</&lt;/g; s/>/&gt;/g;
				$TEXT .= $_;
			}
			close FH;
			chomp $TEXT;
		}elsif($WIKI_NEW_PAGE eq "" || $PAGE !~ m/$WIKI_NEW_PAGE/o){
			exit;
		}
	}elsif($QUERY_STRING =~ m/^wikieditback(?:=([0-9]+))?$/){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?wikieditback	edit the current-1 version of wiki PAGE
# u.cgi/PAGE?wikieditback=([0-9]+) edit the current-\1 version of wiki PAGE
		my $version = get_version($PAGE);
		my $backversion = $version - ($1 eq ""?1:$1);
		{
			open FH, "$PAGE.txt" or die; local $/ = undef;
			$TEXT = <FH>;
			close FH;
		}
		exit unless($TEXT =~ m/^#!wiki\n/);
		if($backversion >= $version || $backversion <= 0){
			close FH;
			print "Content-Type: text/plain\n\n".
				"Current version: $version\n";
			exit;
		}
		{
			open FH, "$PAGE.txt.v"; local $/ = "\x01\n";
			while(<FH>){
				m/^(.*?)\n(.*)\x01\n$/s;
				$TEXT = patch($TEXT, $2);
				last if($backversion == $1);
			}
			close FH;
		}
		exit unless($TEXT =~ m/^#!wiki\n(.*)$/s);
		$TEXT = $1;
		$TEXT =~ s/&/&amp;/g;
		$TEXT =~ s/</&lt;/g;
		$TEXT =~ s/>/&gt;/g;
		chomp $TEXT;
	}else{
		exit;
	}
	local $VERSION = get_version($PAGE) + 1;
	local $TITLE = "Uniqki Wiki Edit: $PAGE";
	print "Content-Type: text/html\n\n";
	do "$TPL/wikiedit.tpl";
	exit;
}elsif($REQUEST_METHOD eq "POST"){
	my %var = get_var();
	exit unless(verify_input($QUERY_STRING, \%var));

	local *FH;
	my $t = time;
	if($QUERY_STRING eq "wikiupload"){
#-------------------------------------------------------------------------------
# wiki upload
		exit if($WIKI_UPLOAD eq "" || !-f "$PAGE.txt" ||
			$var{file} eq "" || $var{file} !~ m/$WIKI_UPLOAD/oi);

		open FH, "$PAGE.txt";
		if(<FH> ne "#!wiki\n"){
			close FH;
			exit;
		}
		close FH;

		mkdir $PAGE, 0755 if(!-d $PAGE);
		open FH, ">$PAGE/$t.$var{file}";
		print FH $var{"file="};
		close FH;
		chmod 0755, "$PAGE/$t.$var{file}" if($HOSTING eq "awardspace");

		print "Content-Type: text/html\n\n";
		do "$TPL/header.tpl";
		(my $f = $var{file}) =~ s/ /%20/g;
		print qq(<pre id="uniqki_uploaded">[$PAGE/$t.$f $var{file}]</pre>\n);
		do "$TPL/footer.tpl";
		exit;
	}
	if(-f "$PAGE.txt"){
		open FH, "$PAGE.txt";
		if(<FH> ne "#!wiki\n"){
			close FH;
			exit;
		}
		close FH;
	}elsif($WIKI_NEW_PAGE eq "" || $PAGE !~ m/$WIKI_NEW_PAGE/o){
		exit;
	}

	local $VERSION = $var{version};
	local $TEXT = $var{text};
	if($VERSION != get_version($PAGE) + 1){
		print "Content-Type: text/html\n\n";
		do "$TPL/updated.tpl";
		exit;
	}
	if($var{file} ne "" && $var{file} =~ m/$WIKI_UPLOAD/oi){
		mkdir $PAGE, 0755 if(!-d $PAGE);
		open FH, ">$PAGE/$t.$var{file}";
		print FH $var{"file="};
		close FH;
		chmod 0755, "$PAGE/$t.$var{file}" if($HOSTING eq "awardspace");
		if($var{preview} eq ""){
			(my $f = $var{file}) =~ s/ /%20/g;
			$TEXT .= "\n[$PAGE/$t.$f $var{file}]";
		}
	}
	if($var{preview} ne ""){
		my $uploaded;
		if($var{file} ne "" && -f "$PAGE/$t.$var{file}"){
			(my $f = $var{file}) =~ s/ /%20/g;
			$uploaded = qq(<pre id="uniqki_uploaded">[$PAGE/$t.$f $var{file}]</pre>\n);
		}

		print "Content-Type: text/html\n\n";
		print make_html($PAGE, 1, $TEXT, $uploaded);
		exit;
	}

	mylock("$PAGE.txt");
	save($PAGE, "#!wiki\n$TEXT\n");
	myunlock("$PAGE.txt");

	make_html($PAGE);

	print "Location: $DOC_BASE/$PAGE.html\n\n";
	exit;
}
	exit;
}

################################################################################
# login, logout, reinstall
if($QUERY_STRING eq "logout"){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?logout		logout
	print "Set-Cookie: uniqki=; domain=$HTTP_HOST; path=$SCRIPT_NAME; ".
		"expires=Tue, 01-Jan-1980 00:00:00 GMT\n";
}else{
	local *FH;
	if($QUERY_STRING ne "reinstall" && open FH, $PW){
		my ($id, $pw);
		if($QUERY_STRING eq "login"){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?login		login
			if($REQUEST_METHOD eq "GET"){
				print "Content-Type: text/html\n\n";
				do "$TPL/login.tpl";
				exit;
			}
			my %var = get_var();
			$id = $var{id};
			$pw = $var{pw};
			$pw = md5_hex("$pw\n");
		}else{
			my $cookie = $HTTP_COOKIE; $cookie =~ s/; /\n/g;
			$cookie =~ m/^uniqki=([^:\n]*):([^:\n]*)$/m;
			$id = $1; $pw = $2;
		}

		my $md5pw = md5_hex("$pw\n");
		my @line = grep /^$id:$md5pw$/, <FH>;
		close FH;
		if($#line >= 0){
			# give me $PW_EXPIRATION days
			my @t = gmtime time + $PW_EXPIRATION * 24 * 60 * 60;
			my @m = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov
				Dec);
			my @w = qw(Sun Mon Tue Wed Thu Fri Sat);
			my $exp = sprintf "%s, %02d-%s-%d %02d:%02d:%02d GMT",
				$w[$t[6]], $t[3], $m[$t[4]], $t[5]+1900, $t[2],
				$t[1], $t[0];
			print "Set-Cookie: uniqki=$id:$pw; ".
				"domain=$HTTP_HOST; path=$SCRIPT_NAME; ".
				"expires=$exp\n";
			$USER = $id;
		}else{
			print "Set-Cookie: uniqki=; domain=$HTTP_HOST; ".
				"path=$SCRIPT_NAME; ".
				"expires=Tue, 01-Jan-1980 00:00:00 GMT\n";
		}
	}elsif($HTTP_HOST ne ""){
################################################################################
# auto install
#-------------------------------------------------------------------------------
# u.cgi/PAGE?reinstall		reinstall, but don't overwrite
		unless(-f $PW){
			open FH, ">$PW";
			# admin id: admin, password: admin
			print FH "admin:a9b86c931199285ac12455814c700982\n";
			close FH;
		}

		unless(-f "u.cfg"){
			open FH, ">u.cfg";
			print FH <<'EOT_UNIQKI';
$TZ = "GMT";
$PW = "u.pw";
$PW_EXPIRATION = 10;
$TPL = "u.tpl";
$CSS = "$TPL/uniqki.css";
$CHARSET = "utf-8";
$RSS_TITLE = "My Uniqki!";
$RSS_DESCRIPTION = $RSS_TITLE;
$HEADER = "";
$FOOTER = "";
$WIKI_HEADER = "";
$WIKI_FOOTER = "";
$WIKI_NEW_PAGE = q();
$WIKI_UPLOAD = q(\.(png|gif|jpg|jpeg|txt|zip)$);
EOT_UNIQKI
			close FH;
		}

################################################################################
# basic template
		mkdir $TPL, 0755 unless(-d $TPL);

		unless(-f "$TPL/header.tpl"){
			open FH, ">$TPL/header.tpl";
			print FH <<'EOT_UNIQKI';
return if($BODY_ONLY);
print <<EOT;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"
	"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
<head>
<title>$TITLE</title>
<meta http-equiv="Content-Type" content="text/html; charset=$CHARSET" />
<link rel="stylesheet" type="text/css" href="$CSS" />
</head>
<body>
EOT
EOT_UNIQKI
			close FH;
		}

		unless(-f "$TPL/footer.tpl"){
			open FH, ">$TPL/footer.tpl";
			print FH <<'EOT_UNIQKI';
return if($BODY_ONLY);
print <<EOT;
</body>
</html>
EOT
EOT_UNIQKI
			close FH;
		}

		unless(-f "$TPL/login.tpl"){
			open FH, ">$TPL/login.tpl";
			print FH <<'EOT_UNIQKI';
do "$TPL/header.tpl";
print <<EOT;
<div id="uniqki_login">
<h1>$PAGE?login</h1>
<form action="$PAGE?login" method="post">
<div>
<input accesskey="i" type="text" id="id" name="id" />
<input accesskey="p" type="password" id="pw" name="pw" />
<input accesskey="l" type="submit" value="login" />
<hr />
<a accesskey="v" href="$DOC_BASE/$PAGE.html">view</a> .
<a accesskey="i" href="$DOC_BASE/index.html">index</a>
</div>
</form>
</div>
EOT
do "$TPL/footer.tpl";
EOT_UNIQKI
			close FH;
		}

		unless(-f "$TPL/admin.tpl"){
			open FH, ">$TPL/admin.tpl";
			print FH <<'EOT_UNIQKI';
do "$TPL/header.tpl";
print qq(<div id="uniqki_admin">
<h1>$PAGE?admin</h1>
<form action="?user" method="post">
<div>
<input accesskey="i" type="text" id="id" name="id" />
<input accesskey="p" type="password" id="pw" name="pw" />
<input type="password" id="pw2" name="pw2" />
<input accesskey="u" type="submit" value="user" />
<ul>
<li>Change password: Enter an existing ID and type new password twice</li>
<li>Add user: Enter new ID and type password twice</li>
<li>Delete user: Enter an existing ID and leave password blank</li>
</ul>
</div>
</form>
<form action="?restore" method="post" enctype="multipart/form-data">
<div>
backup: <a href="$U_CGI?backup">all pages</a>).($PAGE eq "" ? "" :
qq( . <a href="?backup">$PAGE</a>)).qq(
<br />
restore: <input accesskey="f" type="file" id="file" name="file" />
<input accesskey="r" type="submit" value="restore" />
<hr />).($PAGE eq "" ? "" :
qq(<a accesskey="v" href="$DOC_BASE/$PAGE.html">view</a> .)).qq(
<a accesskey="i" href="$DOC_BASE/index.html">index</a>
</div>
</form>
</div>
);
do "$TPL/footer.tpl";
EOT_UNIQKI
			close FH;
		}

		unless(-f "$TPL/display.tpl"){
			open FH, ">$TPL/display.tpl";
			print FH <<'EOT_UNIQKI';
do "$TPL/header.tpl";
my $date = scalar localtime;
print <<EOT;
<div id="uniqki_display">
$TEXT
</div>
<!-- # --><div id="uniqki_menu">
<!-- # --><hr />
<!-- # --><a accesskey="r" href="$U_CGI/$PAGE?refresh">refresh</a> .
<!-- # --><a accesskey="e" href="$U_CGI/$PAGE?edit">edit</a> .
<!-- # --><a accesskey="l" href="$U_CGI/$PAGE?loginout">loginout</a> .
<!-- # --><a accesskey="i" href="index.html">index</a><br />
<!-- # --><small><i>
<!-- # -->$date .
<!-- # --><a href="http://validator.w3.org/check?uri=referer">XHTML 1.1</a> .
<!-- # --><a href="http://jigsaw.w3.org/css-validator/check/referer">CSS</a> .
<!-- # -->Powered by <a href="http://uniqki.ath.cx">Uniqki</a>!
<!-- # --></i></small>
<!-- # --></div>
EOT
do "$TPL/footer.tpl";
EOT_UNIQKI
			close FH;
		}

		unless(-f "$TPL/edit.tpl"){
			open FH, ">$TPL/edit.tpl";
			print FH <<'EOT_UNIQKI';
do "$TPL/header.tpl";
print <<EOT;
<div id="uniqki_edit">
<h1>$PAGE?edit</h1>
<form action="$PAGE?edit" method="post" enctype="multipart/form-data">
<div>
<input type="hidden" id="version" name="version" value="$VERSION" />
<textarea accesskey="e" id="text" name="text" rows="25" cols="80">$TEXT</textarea>
<br />
<input accesskey="p" type="submit" id="preview" name="preview" value="preview" />
<input accesskey="s" type="submit" id="save" name="save" value="save" /> .
upload <input accesskey="u" type="file" id="file" name="file" /> .
<a accesskey="c" href="$DOC_BASE/$PAGE.html">cancel</a>
</div>
</form>
EOT
do "$TPL/footer.tpl";
EOT_UNIQKI
			close FH;
		}

		unless(-f "$TPL/updated.tpl"){
			open FH, ">$TPL/updated.tpl";
			print FH <<'EOT_UNIQKI';
do "$TPL/header.tpl";
print <<EOT;
<div id="uniqki_updated">
<h1>$PAGE updated!</h1>
Please save your changes and read <a href="$DOC_BASE/$PAGE.html">the latest version</a>!
<br />
<textarea accesskey="e" id="text" name="text" rows="25" cols="80">$TEXT</textarea>
</div>
EOT
do "$TPL/footer.tpl";
EOT_UNIQKI
			close FH;
		}

		unless(-f "$TPL/wikidisplay.tpl"){
			open FH, ">$TPL/wikidisplay.tpl";
			print FH <<'EOT_UNIQKI';
do "$TPL/header.tpl";
my $date = scalar localtime;
print <<EOT;
<div id="uniqkiwiki_display">
$TEXT
</div>
<!-- # --><div id="uniqkiwiki_menu">
<!-- # --><a accesskey="e" href="$U_CGI/$PAGE?wikiedit">EditPage</a> .
<!-- # --><a href="$U_CGI/$PAGE?diff=-1">Diff</a> .
<!-- # --><a accesskey="l" href="$U_CGI?search=$PAGE%5C.html&amp;link=1">BackLink</a> .
<!-- # --><a accesskey="i" href="index.html">Index</a><br />
<!-- # --><small><i>
<!-- # -->$date .
<!-- # --><a href="http://validator.w3.org/check?uri=referer">XHTML 1.1</a> .
<!-- # --><a href="http://jigsaw.w3.org/css-validator/check/referer">CSS</a> .
<!-- # -->Powered by <a href="http://uniqki.ath.cx">Uniqki</a>!
<!-- # --></i></small>
<!-- # --></div>
EOT
do "$TPL/footer.tpl";
EOT_UNIQKI
			close FH;
		}

		unless(-f "$TPL/wikiedit.tpl"){
			open FH, ">$TPL/wikiedit.tpl";
			print FH <<'EOT_UNIQKI';
do "$TPL/header.tpl";
print <<EOT;
<div id="uniqkiwiki_edit">
<h1>$PAGE?wikiedit</h1>
<form action="$PAGE?wikiedit" method="post" enctype="multipart/form-data">
<div>
<input type="hidden" id="version" name="version" value="$VERSION" />
<textarea accesskey="e" id="text" name="text" rows="25" cols="80">$TEXT</textarea><br />
<input accesskey="p" type="submit" id="preview" name="preview" value="Preview" />
<input accesskey="s" type="submit" id="save" name="save" value="Save" /> .
EOT
if($WIKI_UPLOAD ne ""){
	print <<EOT;
Upload <input accesskey="u" type="file" id="file" name="file" /> .
EOT
}
print <<EOT;
<a accesskey="c" href="$DOC_BASE/$PAGE.html">Cancel</a>
</div>
</form>
</div>
EOT
do "$TPL/footer.tpl";
EOT_UNIQKI
			close FH;
		}

		unless(-f "$TPL/uniqki.css"){
			open FH, ">$TPL/uniqki.css";
			print FH <<'EOT_UNIQKI';
/******************************************************************************/
form {
	margin:			0px;
}
pre {
	background-color:	#dddddd;
}

/******************************************************************************/
#uniqki_login {
}
#uniqki_admin {
}
#uniqki_edit {
}
#uniqki_updated {
}
#uniqki_uploaded {
	border-bottom:		1px dashed #999999;
	padding-bottom:		20px;
}
#uniqki_display {
}
#uniqki_menu {
}
.uniqki_toc {
	border-bottom:		1px dashed #999999;
	padding-bottom:		20px;
}
.uniqki_toc dl {
	margin-top:		5px;
	margin-bottom:		5px;
}
.uniqki_h7 {
	font-size:		60%;
	font-weight:		bold;
	margin-top:		5px;
	margin-bottom:		5px;
}
.uniqki_table {
	border-collapse:	collapse;
}
.uniqki_table td {
	border:			1px solid #999999;
	padding:		3px;
}

/******************************************************************************/
#uniqkiwiki_display {
	background-color:	#eeeeee;
	color:			#000000;
	border:			1px solid #999999;
	padding:		5px;
}
#uniqkiwiki_menu {
	padding:		5px;
}
#uniqkiwiki_edit {
}

/******************************************************************************/
.goto_input {
}

/******************************************************************************/
#diff {
}
.diff_unchanged {
	font-family:		monospace;
}
.diff_added {
	background-color:	#66cccc;
	color:			#000000;
	font-family:		monospace;
}
.diff_deleted {
	background-color:	#ff99cc;
	color:			#000000;
	font-family:		monospace;
	text-decoration:	line-through;
}
.diff_modified {
	background-color:	#cccccc;
	color:			#000000;
	font-family:		monospace;
}
.diff_modified_added {
	background-color:	#66cccc;
	color:			#000000;
	font-family:		monospace;
}
.diff_modified_deleted {
	background-color:	#ff99cc;
	color:			#000000;
	font-family:		monospace;
	text-decoration:	line-through;
}

/******************************************************************************/
#ls {
}
#ls h1 {
	font-size:		100%;
}
.ls_time {
	font-size:		70%;
	font-style:		italic;
}

/******************************************************************************/
#search {
}
#search h1 {
	font-size:		100%;
}
.search_highlight {
	font-weight:		bold;
}
.search_input {
}

/******************************************************************************/
.comment_input {
	border-top:		1px solid #aaaaaa;
	margin-top:		10px;
	padding-top:		10px;
}
EOT_UNIQKI
			close FH;
		}

		if($QUERY_STRING eq "reinstall"){
			print "Location: $HTTP_BASE$SCRIPT_NAME/$PAGE\n\n";
		}else{
			print "Location: $HTTP_BASE$SCRIPT_NAME/$PAGE?login\n\n";
		}
		exit;
################################################################################
	}
}

################################################################################
# admin actions
if($QUERY_STRING eq "loginout"){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?loginout		loginout
	print "Location: $HTTP_BASE$SCRIPT_NAME/$PAGE?".
		($USER eq "" ? "login" : "logout")."\n\n";
	exit;
}elsif($USER eq ""){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?ACTION		user has to login to perform ACTION
	$PAGE = "index" if($PAGE eq "");
	print "Location: $DOC_BASE/$PAGE.html\n\n";
	exit;
}elsif($QUERY_STRING eq "admin"){
#-------------------------------------------------------------------------------
# u.cgi?admin			admin page
# u.cgi/PAGE?admin		admin page
	print "Content-Type: text/html\n\n";
	do "$TPL/admin.tpl";
	exit;
}elsif($QUERY_STRING eq "user"){
#-------------------------------------------------------------------------------
# u.cgi?user			user management
# u.cgi/PAGE?user		user management
	my %var = get_var();
	if($var{id} ne "" && $var{pw} eq $var{pw2}){
		my $md5pw = md5_hex(md5_hex("$var{pw}\n")."\n");
		my $upw = "";
		my $found = 0;
		local *FH;
		open FH, $PW;
		while(<FH>){
			if(/^$var{id}:/){
				# delete user
				next if($var{pw} eq "");
				$_ = "$var{id}:$md5pw\n";
				$found = 1;
			}
			$upw .= $_;
		}
		$upw .= "$var{id}:$md5pw\n" if(!$found && $var{pw} ne "");
		close FH;

		open FH, ">$PW";
		print FH $upw;
		close FH;
	}
	print "Location: $HTTP_BASE$SCRIPT_NAME/$PAGE?admin\n\n";
	exit;
}elsif($QUERY_STRING eq "backup"){
#-------------------------------------------------------------------------------
# u.cgi?backup			backup all pages
# u.cgi/PAGE?backup		backup PAGE
	backup($PAGE);
	exit;
}elsif($QUERY_STRING eq "restore"){
#-------------------------------------------------------------------------------
# u.cgi?restore			restore
# u.cgi/PAGE?restore		restore
	restore();
	print "Location: $HTTP_BASE$SCRIPT_NAME/$PAGE?admin\n\n";
	exit;
}elsif($QUERY_STRING =~ m/^refresh(?:&.+)?$/){
#-------------------------------------------------------------------------------
# u.cgi?refresh			refresh all
# u.cgi?refresh&glob=GLOB	refresh GLOB pages
	my %var = get_var();
	my $glob = $var{glob};
	my $_pre_text = $pre_text;
	my $_line = $line;
	my $_end = $end;
	my $_post_text = $post_text;

	print "Content-Type: text/html\n\n";
	do "$TPL/header.tpl";
	print qq(<div id="ls">\n<h1>refresh ).
		($glob eq "" ? "all" : $glob).qq(</h1>\n);
	foreach($glob eq "" ? (<.*.txt>, <*.txt>) : <$glob.txt>){
		next if(index($_, "/") >= 0 || !-f $_);
		s/\.txt$//;
		$PAGE = $_;
		$pre_text = $_pre_text;
		$line = $_line;
		$end = $_end;
		$post_text = $_post_text;
		make_html($PAGE);
		print qq(<div><a href="$PAGE.html">$PAGE</a></div>\n);
	}
	print qq(</div>\n);
	do "$TPL/footer.tpl";
	exit;
}elsif($PAGE eq ""){
#-------------------------------------------------------------------------------
# u.cgi?ACTION			index page
	print "Location: $HTTP_BASE$SCRIPT_NAME/index\n\n";
	exit;
}elsif($QUERY_STRING eq "delete" && $FILE ne ""){
#-------------------------------------------------------------------------------
# u.cgi/PAGE/FILE?delete	delete PAGE/FILE
	my $dir = "";
	if(unlink "$PAGE/$FILE"){
		$dir = "$PAGE/$FILE"; $dir =~ s#/[^/]+$##;
	}elsif(-d "$PAGE/$FILE"){
		$dir = "$PAGE/$FILE";
	}
	if($dir ne ""){
		while(index($dir, "/") >= 0){
			rmdir $dir;
			$dir =~ s#/[^/]*$##;
		}
		rmdir $dir;
	}
	$rebuild = 1;
}elsif($QUERY_STRING eq "upload"){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?upload		upload PAGE/FILE
	my %var = get_var();
	exit unless(verify_input("upload", \%var));

	if($var{file} ne ""){
		local *FH;
		mkdir $PAGE, 0755 if(!-d $PAGE);
		open FH, ">$PAGE/$var{file}";
		print FH $var{"file="};
		close FH;
		chmod 0755, "$PAGE/$var{file}" if($HOSTING eq "awardspace");
		$rebuild = 1;
	}
}elsif($QUERY_STRING =~ m/^edit(?:=(-?[0-9]+))?$/ ||
	($QUERY_STRING ne "delete" && !-f "$PAGE.txt")){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?edit		create/edit PAGE
# u.cgi/PAGE?edit=([0-9]+)	edit the version \1 of PAGE
# u.cgi/PAGE?edit=-([0-9]+)	edit the current-\1 version of PAGE
	local *FH;
	local ($VERSION, $TEXT);
	local $TITLE;
	if($REQUEST_METHOD eq "GET" || $QUERY_STRING eq "login"){
		$VERSION = get_version($PAGE);
		my $version = ($1 > 0 ? $1 : $VERSION + $1);
		$TEXT = "";
		if(open FH, "$PAGE.txt"){
			local $/ = undef;
			$TEXT = <FH>;
			close FH;

			if($version > 0 && $version < $VERSION){
				open FH, "$PAGE.txt.v"; local $/ = "\x01\n";
				while(<FH>){
					m/^(.*?)\n(.*)\x01\n$/s;
					$TEXT = patch($TEXT, $2);
					last if($version == $1);
				}
				close FH;
			}
		}
		$VERSION++;
		if($TEXT ne ""){
			$TEXT =~ s/&/&amp;/g;
			$TEXT =~ s/</&lt;/g;
			$TEXT =~ s/>/&gt;/g;
			chomp $TEXT;
		}
		$TITLE = "Uniqki Edit: $PAGE";
		print "Content-Type: text/html\n\n";
		do "$TPL/edit.tpl";
		exit;
	}

	my %var = get_var();
	exit unless(verify_input("edit", \%var));

	$VERSION = $var{version};
	$TEXT = $var{text};
	if($VERSION != get_version($PAGE) + 1){
		$TITLE = "Uniqki Updated: $PAGE";
		print "Content-Type: text/html\n\n";
		do "$TPL/updated.tpl";
		exit;
	}
	if($var{file} ne ""){
		mkdir $PAGE, 0755 if(!-d $PAGE);
		open FH, ">$PAGE/$var{file}";
		print FH $var{"file="};
		close FH;
		chmod 0755, "$PAGE/$var{file}" if($HOSTING eq "awardspace");
	}
	if($var{preview} ne ""){
		my $uploaded;
		if($var{file} ne "" && -f "$PAGE/$var{file}"){
			(my $f = $var{file}) =~ s/ /%20/g;
			$uploaded = qq(<pre id="uniqki_uploaded">[$PAGE/$f $var{file}]</pre>\n);
		}

		print "Content-Type: text/html\n\n";
		print make_html($PAGE, 0, $TEXT, $uploaded);
		exit;
	}
	mylock("$PAGE.txt");
	save($PAGE, "$TEXT\n");
	myunlock("$PAGE.txt");
}elsif($QUERY_STRING =~ m/^revert(?:=(-?[0-9]+))?$/){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?revert		go back to the previous version of PAGE
# u.cgi/PAGE?revert=([0-9]+)	go back to the \1 version of PAGE
# u.cgi/PAGE?revert=-([0-9]+)	go back to the current-\1 version of PAGE
	local *FH;
	mylock("$PAGE.txt");
	my $current_version = get_version($PAGE);
	my $version = ($1 > 0 ? $1 : $current_version + ($1 < 0 ? $1 : -1));
	if($version > 0 && $version < $current_version){
		open FH, "$PAGE.txt"; local $/ = undef;
		my $text = <FH>;
		close FH;
		open FH, "$PAGE.txt.v"; local $/ = "\x01\n";
		while(<FH>){
			m/^(.*?)\n(.*)\x01\n$/s;
			$text = patch($text, $2);
			if($version == $1){
				$rebuild = 1;
				last;
			}
		}
		if($rebuild){
			local $/ = undef;
			my $txtv = <FH>;
			close FH;
			open FH, ">$PAGE.txt";
			print FH $text;
			close FH;
			if($txtv eq ""){
				unlink "$PAGE.txt.v";
			}else{
				open FH, ">$PAGE.txt.v";
				print FH $txtv;
				close FH;
			}
		}else{
			close FH;
		}
	}
	myunlock("$PAGE.txt");
}elsif($QUERY_STRING eq "delete"){
#-------------------------------------------------------------------------------
# u.cgi/PAGE?delete		delete PAGE
	unlink "$PAGE.txt", "$PAGE.txt.v", "$PAGE.html";
	rmrf($PAGE);
	$PAGE = "index";
}

#-------------------------------------------------------------------------------
# rebuild, if requested, and redirect
make_html($PAGE) if($rebuild);

print "Location: $DOC_BASE/$PAGE.html\n\n";
